# Logger names
TRAIN_LOGGER_NAME = 'train'
HYPEROPT_LOGGER_NAME = 'hyperparameter-optimization'

# Save file names
MODEL_FILE_NAME = 'model.pt'
TEST_SCORES_FILE_NAME = 'test_scores.csv'
HYPEROPT_SEED_FILE_NAME = 'hyperopt_seeds.txt'
